@@ -3,7 +3,7 @@
 #
 #  (C) Paul Evans, 2021 -- leonerd@leonerd.org.uk
 
-package XS::Parse::Infix 0.14;
+package XS::Parse::Infix 0.15;
 
 use v5.14;
 use warnings;
@@ -20,7 +20,7 @@ C<XS::Parse::Infix> - XS functions to assist in parsing infix operators
 This module provides some XS functions to assist in writing syntax modules
 that provide new infix operators as perl syntax, primarily for authors of
 syntax plugins. It is unlikely to be of much use to anyone else; and highly
-unlikely to be of any use when writing prl code using these. Unless you are
+unlikely to be of any use when writing perl code using these. Unless you are
 writing a syntax plugin using XS, this module is not for you.
 
 This module is also currently experimental, and the design is still evolving
@@ -112,6 +112,8 @@ used at various stages of parsing.
    struct XSParseInfixHooks {
       enum XSParseInfixClassification cls;
 
+      const char *wrapper_func_name;
+
       const char *permit_hintkey;
       bool (*permit)(pTHX_ void *hookdata);
 
@@ -154,6 +156,43 @@ optree fragment that consumes the LHS and RHS arguments to implement the
 semantics of the operator. If this is not present, then the C<ppaddr> will be
 used instead to construct a new BINOP of the C<OP_CUSTOM> type.
 
+=head2 The Wrapper Function
+
+Additionally, if the C<wrapper_func_name> field is set to a string, this gives
+the (fully-qualified) name for a function to be generated as part of
+registering the operator. This newly-generated function will act as a wrapper
+for the operator, which is assumed to take two simple scalar arguments. The
+result of invoking the function on those arguments will be determined by using
+the operator code.
+
+   $result = $lhs OP $rhs;
+
+   $result = WRAPPERFUNC( $lhs, $rhs );
+
+This creates a convenience for allowing perls that do not support
+C<PL_infix_plugin> to still have access to the operator.
+
+The wrapper function also includes a call-checker, which attempts to inline
+the operator directly into the callsite. Thus, in simple cases where the
+function is called directly in cases such as the following, no C<ENTERSUB>
+overhead will be incurred, and the generated optree will be identical to that
+which would have been generated by using infix operator syntax directly, even
+on versions of perl that do not have C<PL_infix_plugin>:
+
+   WRAPPERFUNC( $lhs, $rhs );
+   WRAPPERFUNC( $lhs, CONSTANT );
+   WRAPPERFUNC( $args[0], $args[1] );
+   WRAPPERFUNC( $lhs, scalar otherfunc() );
+
+The checker is very pessimistic and will only rewrite callsites where it
+determines this can be done safely. It will not rewrite any of the following
+forms:
+
+   WRAPPERFUNC( $onearg );            # not enough args
+   WRAPPERFUNC( $x, $y, $z );         # too many args
+   WRAPPERFUNC( @args[0,1] );         # not a scalar
+   WRAPPERFUNC( $lhs, otherfunc() );  # not a scalar
+
 =cut
 
 =head1 AUTHOR
