@@ -15,12 +15,38 @@
 #define HAVE_PERL_VERSION(R, V, S) \
     (PERL_REVISION > (R) || (PERL_REVISION == (R) && (PERL_VERSION > (V) || (PERL_VERSION == (V) && (PERL_SUBVERSION >= (S))))))
 
+#include "make_argcheck_ops.c.inc"
 #include "newOP_CUSTOM.c.inc"
 
 #if HAVE_PERL_VERSION(5,32,0)
 #  define HAVE_OP_ISA
 #endif
 
+/* These only became full API macros at perl v5.22, but they're available as
+ * the full Perl_... name before that
+ */
+#ifndef block_start
+#  define block_start(a)         Perl_block_start(aTHX_ a)
+#endif
+
+#ifndef block_end
+#  define block_end(a,b)         Perl_block_end(aTHX_ a,b)
+#endif
+
+
+#ifndef OpSIBLING
+#  define OpSIBLING(op)  (op->op_sibling)
+#endif
+
+#ifndef OpMORESIB_set
+#  define OpMORESIB_set(op,sib)  ((op)->op_sibling = (sib))
+#endif
+
+struct HooksAndData {
+  const struct XSParseInfixHooks *hooks;
+  void *data;
+};
+
 struct Registration;
 struct Registration {
 #ifdef HAVE_PL_INFIX_PLUGIN
@@ -33,20 +59,19 @@ struct Registration {
   STRLEN      oplen;
   enum XSParseInfixClassification cls;
 
-  const struct XSParseInfixHooks *hooks;
-  void *hookdata;
+  struct HooksAndData hd;
 
   STRLEN permit_hintkey_len;
 };
 
 static struct Registration *registrations;
 
-static OP *new_op(pTHX_ const struct XSParseInfixHooks *hooks, U32 flags, OP *lhs, OP *rhs, void *hookdata)
+static OP *new_op(pTHX_ const struct HooksAndData hd, U32 flags, OP *lhs, OP *rhs)
 {
-  if(hooks->new_op)
-    return (*hooks->new_op)(aTHX_ flags, lhs, rhs, hookdata);
+  if(hd.hooks->new_op)
+    return (*hd.hooks->new_op)(aTHX_ flags, lhs, rhs, hd.data);
 
-  OP *ret = newBINOP_CUSTOM(hooks->ppaddr, flags, lhs, rhs);
+  OP *ret = newBINOP_CUSTOM(hd.hooks->ppaddr, flags, lhs, rhs);
 
   /* TODO: opchecker? */
 
@@ -63,7 +88,7 @@ OP *parse(pTHX_ OP *lhs, struct Perl_custom_infix *def)
   lex_read_space(0);
   OP *rhs = parse_termexpr(0);
 
-  return new_op(aTHX_ reg->hooks, 0, lhs, rhs, reg->hookdata);
+  return new_op(aTHX_ reg->hd, 0, lhs, rhs);
 }
 
 static STRLEN (*next_infix_plugin)(pTHX_ char *, STRLEN, struct Perl_custom_infix **);
@@ -78,18 +103,18 @@ static STRLEN my_infix_plugin(pTHX_ char *op, STRLEN oplen, struct Perl_custom_i
   struct Registration *reg;
   for(reg = registrations; reg; reg = reg->next) {
     /* custom registrations have hooks, builtin ones do not */
-    if(!reg->hooks)
+    if(!reg->hd.hooks)
       continue;
 
     if(reg->oplen != oplen || !strEQ(reg->info.opname, op))
       continue;
 
-    if(reg->hooks->permit_hintkey &&
-      (!hints || !hv_fetch(hints, reg->hooks->permit_hintkey, reg->permit_hintkey_len, 0)))
+    if(reg->hd.hooks->permit_hintkey &&
+      (!hints || !hv_fetch(hints, reg->hd.hooks->permit_hintkey, reg->permit_hintkey_len, 0)))
       continue;
 
-    if(reg->hooks->permit &&
-      !(*reg->hooks->permit)(aTHX_ reg->hookdata))
+    if(reg->hd.hooks->permit &&
+      !(*reg->hd.hooks->permit)(aTHX_ reg->hd.data))
       continue;
 
     *def = &reg->def;
@@ -136,12 +161,12 @@ bool XSParseInfix_parse(pTHX_ enum XSParseInfixSelection select, struct XSParseI
     if(!(selection & (1 << reg->cls)))
       continue;
 
-    if(reg->hooks && reg->hooks->permit_hintkey &&
-      (!hints || !hv_fetch(hints, reg->hooks->permit_hintkey, reg->permit_hintkey_len, 0)))
+    if(reg->hd.hooks && reg->hd.hooks->permit_hintkey &&
+      (!hints || !hv_fetch(hints, reg->hd.hooks->permit_hintkey, reg->permit_hintkey_len, 0)))
       continue;
 
-    if(reg->hooks && reg->hooks->permit &&
-      !(*reg->hooks->permit)(aTHX_ reg->hookdata))
+    if(reg->hd.hooks && reg->hd.hooks->permit &&
+      !(*reg->hd.hooks->permit)(aTHX_ reg->hd.data))
       continue;
 
     *infop = &reg->info;
@@ -156,11 +181,92 @@ bool XSParseInfix_parse(pTHX_ enum XSParseInfixSelection select, struct XSParseI
 OP *XSParseInfix_new_op(pTHX_ const struct XSParseInfixInfo *info, U32 flags, OP *lhs, OP *rhs)
 {
   if(info->opcode == OP_CUSTOM)
-    return new_op(aTHX_ info->hooks, flags, lhs, rhs, info->hookdata);
+    return new_op(aTHX_ (struct HooksAndData) {
+        .hooks = info->hooks,
+        .data  = info->hookdata,
+      }, flags, lhs, rhs);
 
   return newBINOP(info->opcode, flags, lhs, rhs);
 }
 
+static OP *ckcall_wrapper_func(pTHX_ OP *op, GV *namegv, SV *ckobj)
+{
+  struct HooksAndData *hd = NUM2PTR(struct HooksAndData *, SvUV(ckobj));
+
+  assert(op->op_type == OP_ENTERSUB);
+
+  /* Attempt to extract the LHS and RHS operands, if we can find them */
+
+  OP *kid = cUNOPx(op)->op_first;
+  /* The first kid is usually an ex-list whose ->op_first begins the actual args list */
+  if(kid->op_type == OP_NULL && kid->op_targ == OP_LIST)
+    kid = cUNOPx(kid)->op_first;
+
+  assert(kid->op_type == OP_PUSHMARK);
+  OP *pushmark = kid;
+
+  OP *left = OpSIBLING(kid);
+  if(!left)
+    return op;
+  if(OP_GIMME(left, 0) != G_SCALAR)
+    return op;
+
+  OP *right = OpSIBLING(left);
+  if(!right)
+    return op;
+  if(OP_GIMME(right, 0) != G_SCALAR)
+    return op;
+
+  kid = OpSIBLING(right);
+  if(!kid)
+    return op;
+  if(OpSIBLING(kid))
+    return op;
+
+  /* Check that kid is now OP_NULL[ OP_GV ] */
+  if(kid->op_type != OP_NULL || kid->op_targ != OP_RV2CV)
+    return op;
+  if(cUNOPx(kid)->op_first->op_type != OP_GV)
+    return op;
+
+  /* Splice out these two args and throw away the old optree */
+  OpMORESIB_set(left, NULL);
+  OpMORESIB_set(right, NULL);
+  OpMORESIB_set(pushmark, kid);
+  op_free(op);
+
+  /* Now build a new optree */
+  op = new_op(aTHX_ *hd, 0, left, right);
+  return op;
+}
+
+static void make_wrapper_func(pTHX_ const struct HooksAndData *hd)
+{
+  /* Prepare to make a new optree-based CV */
+  I32 floor_ix = start_subparse(FALSE, 0);
+  SAVEFREESV(PL_compcv);
+
+  SV *funcname = newSVpvn(hd->hooks->wrapper_func_name, strlen(hd->hooks->wrapper_func_name));
+
+  I32 save_ix = block_start(TRUE);
+
+  OP *body = NULL;
+
+  body = op_append_list(OP_LINESEQ, body,
+      make_argcheck_ops(2, 0, 0, funcname));
+
+  /* Body of the function is just  shift() OP shift() */
+  body = op_append_list(OP_LINESEQ, body,
+      new_op(aTHX_ *hd, 0, newOP(OP_SHIFT, OPf_SPECIAL), newOP(OP_SHIFT, OPf_SPECIAL)));
+
+  SvREFCNT_inc(PL_compcv);
+  body = block_end(save_ix, body);
+
+  CV *cv = newATTRSUB(floor_ix, newSVOP(OP_CONST, 0, funcname), NULL, NULL, body);
+
+  cv_set_call_checker(cv, &ckcall_wrapper_func, newSVuv(PTR2UV(hd)));
+}
+
 static void reg_builtin(pTHX_ const char *opname, enum XSParseInfixClassification cls, OPCODE opcode)
 {
   struct Registration *reg;
@@ -173,8 +279,8 @@ static void reg_builtin(pTHX_ const char *opname, enum XSParseInfixClassificatio
   reg->oplen  = strlen(opname);
   reg->cls    = cls;
 
-  reg->hooks    = NULL;
-  reg->hookdata = NULL;
+  reg->hd.hooks = NULL;
+  reg->hd.data  = NULL;
 
   reg->permit_hintkey_len = 0;
 
@@ -201,8 +307,8 @@ void XSParseInfix_register(pTHX_ const char *opname, const struct XSParseInfixHo
   reg->oplen  = strlen(opname);
   reg->cls    = hooks->cls;
 
-  reg->hooks    = hooks;
-  reg->hookdata = hookdata;
+  reg->hd.hooks = hooks;
+  reg->hd.data  = hookdata;
 
   if(hooks->permit_hintkey)
     reg->permit_hintkey_len = strlen(hooks->permit_hintkey);
@@ -213,6 +319,9 @@ void XSParseInfix_register(pTHX_ const char *opname, const struct XSParseInfixHo
     reg->next = registrations;
     registrations = reg;
   }
+
+  if(hooks->wrapper_func_name)
+    make_wrapper_func(aTHX_ &reg->hd);
 }
 
 void XSParseInfix_boot(pTHX)
